from __future__ import annotations
from typing import Any

import Cython.Build
from setuptools.extension import Extension


def build(setup_kwargs: dict[str, Any]) -> None:
    cython_modules = Cython.Build.cythonize(
        [
            Extension(
                "deqr.quirc",
                sources=[
                    "deqr/quirc.pyx",
                    "deps/quirc/lib/decode.c",
                    "deps/quirc/lib/identify.c",
                    "deps/quirc/lib/quirc.c",
                    "deps/quirc/lib/version_db.c",
                ],
                include_dirs=["deps/quirc/lib"],
                extra_compile_args=["-fdiagnostics-color=always"],
            ),
            Extension(
                "deqr.qrdec",
                sources=[
                    "deqr/qrdec.pyx",
                    "deps/qrdec/src/bch15_5.c",
                    "deps/qrdec/src/isaac.c",
                    "deps/qrdec/src/qrdec.c",
                    "deps/qrdec/src/rs.c",
                    "deps/qrdec/src/util.c",
                ],
                include_dirs=["deps/qrdec/src"],
                extra_compile_args=["-fdiagnostics-color=always"],
            ),
            Extension(
                "deqr.binarize",
                sources=["deqr/binarize.pyx", "deps/qrdec/src/binarize.c"],
                include_dirs=["deps/qrdec/src"],
                extra_compile_args=["-fdiagnostics-color=always"],
            ),
        ],
        language_level=3,
        compiler_directives={"embedsignature": True},
    )

    setup_kwargs.update(
        {
            "ext_modules": cython_modules,
            "zip_safe": False,
        }
    )



# Copyright (C) 2021 torque <torque@users.noreply.github.com>

# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.

# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from typing import Optional

cimport cython

import enum

from . cimport binarize as bnz, quircdecl
from . import datatypes, image

from libc.string cimport memcpy
from libc.stdio cimport printf, fopen, fwrite, fclose, FILE


cdef class QuircDecoder:
    """
    QR code decoder using the Quirc backend.

    :raises MemoryError: if the QR code reader context allocation fails.
    """

    cdef quircdecl.quirc *_chndl

    def __cinit__(self):
        self._chndl = quircdecl.quirc_new()
        if self._chndl is NULL:
            raise MemoryError

    def __dealloc__(self):
        if self._chndl is not NULL:
            quircdecl.quirc_destroy(self._chndl)


    def decode(
        self,
        image_data,
        binarize: bool = True,
        binarize_invert: bool = False,
        convert_data: bool = True,
        byte_charset: Optional[str] = "utf-8"
    ):
        """
        Decode all detectable QR codes in an image.

        .. warning::

            Binarization is done in-place on the image data buffer and
            changes the data. While this operation should be idempotent, it
            mutates the input data, which is a side effect.

            Additionally, inverting during binarization is also done in-place on
            the image data buffer and is *not* idempotent, so if image data is
            reused, it will be re-inverted and not decode correctly the second
            time.

        :param image_data:
            A python object containing the pixel data of the image to search for
            QR codes. This can be any format supported
            by :class:`image.ImageLoader`.

        :param binarize:
            If ``True``, binarize the input image (i.e. convert all
            pixels to either fully black or fully white). The decoder is
            unlikely to work properly on images that have not been binarized.

        :param binarize_invert:
            If ``True``, binarization inverts the reflectance (i.e dark pixels
            become white and light pixels become black).

        :param convert_data:
            If ``True``, all data entry payloads are  decoded into
            "native" python types:

            - :py:obj:`deqr.datatypes.QRDataType.NUMERIC` => :py:class:`int`,
            - :py:obj:`deqr.datatypes.QRDataType.ALPHANUMERIC` =>
              :py:class:`str` as ``ascii``
            - :py:obj:`deqr.datatypes.QRDataType.KANJI` =>
              :py:class:`str` as ``shift-jis``
            - :py:obj:`deqr.datatypes.QRDataType.BYTE` =>
              :py:class:`str` using the ``byte_charset`` parameter

            If ``False``, no conversion will be done and all data entry data
            will be returned as :py:class:`bytes`.

        :param byte_charset:
            The charset to use for converting all decoded data entries of type
            :py:obj:`deqr.datatypes.QRDataType.BYTE` found in the image. If ``None``,
            then data entries of type :py:obj:`deqr.datatypes.QRDataType.BYTE` will not be
            converted, even if ``convert_data`` is ``True``.

        :return: A list of objects containing information from the decoded QR
            codes.

        :raises MemoryError: if the decoder image buffer allocation fails.
        :raises TypeError: if ``image_data`` is of an unsupported type.
        :raises ValueError: if ``image_data`` is malformed somehow.
        """

        cdef int idx = 0
        cdef quircdecl.quirc_code code
        cdef quircdecl.quirc_data data

        decoded: list[datatypes.QRCode] = []

        if not isinstance(image_data, image.ImageLoader):
            image_data = image.ImageLoader(image_data)

        if binarize:
            bnz.binarize(
                image_data.data,
                image_data.width,
                image_data.height,
                binarize_invert
            )

        if convert_data:
            if byte_charset is not None:
                byte_converter = lambda d: d.decode(byte_charset)
            else:
                byte_converter = lambda d: d

            converters = {
                datatypes.QRDataType.NUMERIC: int,
                datatypes.QRDataType.ALPHANUMERIC: lambda d: d.decode("ascii"),
                datatypes.QRDataType.KANJI: lambda d: d.decode("shift-jis"),
                datatypes.QRDataType.BYTE: byte_converter,
            }

        for idx in range(
            self._set_image(
                image_data.data, image_data.width, image_data.height
            )
        ):
            quircdecl.quirc_extract(self._chndl, idx, &code)
            if quircdecl.quirc_decode(&code, &data) != quircdecl.QUIRC_SUCCESS:
                continue

            data_entries = (
                datatypes.QRCodeData(data.data_type, data.payload[:data.payload_len]),
            )

            if convert_data:
                for entry in data_entries:
                    entry.data = converters[entry.type](entry.data)

            center = self.compute_center_from_bounds(code.corners)

            decoded.append(
                datatypes.QRCode(
                    version=data.version,
                    ecc_level=data.ecc_level,
                    mask=data.mask,
                    data_entries=data_entries,
                    corners=tuple((c.x, c.y) for c in code.corners),
                    center=(center.x, center.y)
                )
            )

        return decoded

    cdef quircdecl.quirc_point compute_center_from_bounds(self, quircdecl.quirc_point corners[4]) nogil:
        cdef int divisor = 0
        cdef int lcoeff = 0
        cdef int rcoeff = 0

        divisor = (
            (corners[0].x - corners[2].x) * (corners[1].y - corners[3].y)
            - (corners[0].y - corners[2].y) * (corners[1].x - corners[3].x)
        )

        if divisor == 0:
            return quircdecl.quirc_point(-1, -1)
        else:
            lcoeff = corners[0].x * corners[2].y - corners[0].y * corners[2].x
            rcoeff = corners[1].x * corners[3].y - corners[1].y * corners[3].x

            return quircdecl.quirc_point(
                (
                    (corners[1].x - corners[3].x) * lcoeff
                    - (corners[0].x - corners[2].x) * rcoeff
                ) // divisor,
                (
                    (corners[1].y - corners[3].y) * lcoeff
                    - (corners[0].y - corners[2].y) * rcoeff
                ) // divisor
            )

    cdef int _set_image(self, quircdecl.uint8[::1] image, int width, int height) nogil:
        if quircdecl.quirc_resize(self._chndl, width, height) != quircdecl.QUIRC_SUCCESS:
            raise MemoryError("could not resize")

        cdef int bufsize = width*height

        cdef unsigned char *buffer = quircdecl.quirc_begin(self._chndl, NULL, NULL)
        memcpy(buffer, &image[0], bufsize)

        quircdecl.quirc_end(self._chndl)
        return quircdecl.quirc_count(self._chndl)



from __future__ import annotations
from typing import Any, Union


class ImageLoader:
    """
    Load image data from common Python data containers.

    This is responsible for manipulating image data into a format that the QR
    decoders can understand. Images are converted to 8-bytes-per-pixel
    grayscale and ordered as a 1-dimensional row-major array of data.

    To avoid unnecessary hard dependencies on large external libraries, the
    source format of provided data is inferred by inspecting the object's type
    description. Supported source formats include the :class:`numpy.ndarray`
    used by numpy and OpenCV, the :class:`cv2.UMat` format from OpenCV (this is
    converted to a :class:`numpy.ndarray`), the :class:`PIL.Image.Image` format
    used by Pillow, or, if all else fails, a
    :class:`tuple[bytes, tuple[int, int]]` containing a pre-converted bytes
    object and its associated width, height pair.

    .. warning::

        It is strongly discouraged to access any of the properties of this class
        after it has been created. :attr:`ImageLoader.data` uses a different
        backing class depending on the source type (:class:`numpy.ndarray` when
        coming from numpy or OpenCV, :class:`bytearray` otherwise) and the QR
        code decoders may do in-place manipulation of the data (e.g. the
        binarization routine manipulates the data in place).
    """

    #: The width, in pixels, of the image data stored in :attr:`data`
    width: int
    #: The height, in pixels, of the image data stored in :attr:`data`
    height: int
    #: The image binary data.
    data: Union[bytearray, numpy.ndarray]

    def __init__(self, data: Any):
        try:
            if "numpy.ndarray" in str(type(data)):
                self._set_data_from_numpy_ndarray(data)
            elif "cv2.UMat" in str(type(data)):
                self._set_data_from_numpy_ndarray(data.get())
            elif "PIL." in str(type(data)):
                self._set_data_from_pil_image(data)
            elif isinstance(data, tuple):
                self._set_data_from_tuple(data)
            else:
                raise Exception
        except (ValueError, TypeError):
            raise
        except Exception:
            raise TypeError(f"data {type(data)} is an incompatible type.")

    def _set_data_from_numpy_ndarray(self, data: numpy.ndarray) -> None:
        if data.dtype.name != "uint8":
            # data.astype is not really usable here as it just does blind casts,
            # which doesn't scale properly from larger types. At the same time,
            # it's not possible to really guess what the appropriate range of
            # the input data is for useful conversion
            raise ValueError(f"np.ndarray dtype {data.dtype} must be uint8")
        if data.ndim == 3:
            # try to convert sanely by averaging values
            data = data.mean(axis=2).astype(data.dtype)
        elif data.ndim != 2:
            raise ValueError(f"np.ndarray doesn't look like a 2D image {data.shape}")

        self.data = data.ravel(order="C")
        self.height, self.width = data.shape

    def _set_data_from_pil_image(self, data):
        if data.mode != "L":
            data = data.convert("L")

        self.data = bytearray(data.tobytes())
        self.width, self.height = data.size

    def _set_data_from_tuple(self, data):
        if (
            len(data) != 2
            or not isinstance(data[0], (bytes, bytearray))
            or len(data[1]) != 2
            or not isinstance(data[1][0], int)
            or not isinstance(data[1][1], int)
        ):
            raise TypeError("Data tuple is not of format (bytes, (int, int))")

        if len(data[0]) != (data[1][0] * data[1][1]):
            raise ValueError(
                f"Wrong amount of image data {len(data[0])} for "
                f"{data[1][0]} by {data[1][1]} image"
            )

        self.data = bytearray(data[0])
        self.width, self.height = data[1]



__version__ = "0.2.0"

# this is to handle local development installs, where the version replacement
# gets reverted after the installation finishes.
if __version__ == "__PLACEHOLDER__":
    from importlib import metadata

    __version__ = metadata.version(__name__)


from . import image, datatypes
from .datatypes import QREccLevel, QRDataType, QRCodeData, QRCode
from .quirc import QuircDecoder
from .qrdec import QRdecDecoder



from __future__ import annotations

import enum


class QREccLevel(enum.IntEnum):
    """
    An :py:class:`enum.IntEnum` representing the QR code's error correction
    code (ECC) level.

    This value represents how much error correction data the QR code contains,
    which directly maps to how much of the code can be missing or occluded
    while still permitting successful decoding.

    The integer value matches the number encoded into the QR code itself, rather
    than corresponding to the amount of ECC data contained. :obj:`QREccLevel.L`
    contains the least data, followed by :obj:`QREccLevel.M`,
    :obj:`QREccLevel.Q` and :obj:`QREccLevel.H` in order of increasing amount of
    ECC data.

    See Wikipedia__ for more information

    .. __: https://en.wikipedia.org/wiki/QR_code#Error_correction
    """

    M = 0  #: The code can successfully decode with 15% of codewords missing.
    L = 1  #: The code can successfully decode with 7% of codewords missing.
    H = 2  #: The code can successfully decode with 30% of codewords missing.
    Q = 3  #: The code can successfully decode with 25% of codewords missing.


class QRDataType(enum.IntEnum):
    """
    An :py:class:`enum.IntEnum` representing the data storage format of a given
    data section.

    QR codes can encode multiple different types of data with varying levels of
    efficiency. An individual QR code can also contain multiple different data
    sections, each with their own storage format.

    See Wikipedia__ for more information.

    .. __: https://en.wikipedia.org/wiki/QR_code#Encoding
    """

    NUMERIC = 1  #: The data was stored as numeric (characters [0-9])
    ALPHANUMERIC = 2  #: The data was stored as alphanumeric ([0-9A-Z] and some)
    BYTE = 4  #: The data was stored as 8-bit bytes (normally a UTF-8 string)
    KANJI = 8  #: the data was stored as 13-bit kanji (Shift-JIS encoded)


class QRCodeData:
    """
    A data segment read from the QR code.
    """

    #: the type of QR code data, which also represents its storage format in the
    #: code, for types that contain data.
    #:
    #: The QR code decoder backend is responsible for decoding the data into a
    #: meaningful byte sequence before providing it to the user, so this is
    #: mostly just metadata about how the data was stored in the QR code.
    #: See :meth:`deqr.QRdecDecoder.decode` for details.
    type: QRDataType
    #: The binary data contained in the segment.
    data: bytes

    def __init__(self, type, data):
        self.type = QRDataType(type)
        self.data = data

    def __str__(self):
        return "QrCodeData(type=%s, data=%s)" % (
            self.type,
            self.data,
        )

    def __repr__(self):
        return str(self)


class QRCode:
    """
    A structured collection of information about a QR code that has been decoded.

    This class cannot usefully be instantiated by the user, but instances of it
    are created by :meth:`deqr.QRdecDecoder.decode`
    or :meth:`deqr.QuircDecoder.decode`.
    """

    #: The QR code version, which determines its size and data storage capacity.
    version: int
    #: The error correction code level of this QR code.
    ecc_level: QREccLevel
    #: The data masking pattern used on the code.
    mask: int

    #: The data segments contained in this QR code.
    #:
    #: deqr currently only produces data entries from segments that contain
    #: data, i.e. entries stored as :obj:`QRDataType.NUMERIC`,
    #: :obj:`QRDataType.ALPHANUMERIC`, :obj:`QRDataType.BYTE`, or
    #: :obj:`QRDataType.KANJI`
    data_entries: tuple[QRCodeData, ...]

    #: The pixel coordinates of the geometric center of the QR code, relative to
    #: the top-left of the image. Coordinates are ordered ``(x, y)``, i.e.
    #: ``(horizontal, vertical)``.
    center: tuple[int, int]

    #: A sequence of pixel coordinates of the four corners of the QR code
    #: relative to the top-left of the image.
    #:
    #: These coordinates always start with the corner containing the top-left
    #: finder pattern (relative to nominal "normal" QR code orientation). For
    #: example, if the QR code is upside down in the image, the first corner in
    #: this list of coordinates will actually be the bottom-right corner of the
    #: code's position within the image.
    corners: tuple[tuple[int, int], ...]

    def __init__(self, version, ecc_level, mask, data_entries, corners, center):
        self.version = version
        self.ecc_level = QREccLevel(ecc_level)
        self.mask = mask
        self.data_entries = data_entries

        self.corners = corners
        self.center = center

    def __str__(self):
        return (
            "QRCode(version=%d, ecc_level=%s, mask=%d, data_entries=%s, "
            "corners=%s, center=%s)"
        ) % (
            self.version,
            self.ecc_level,
            self.mask,
            self.data_entries,
            self.corners,
            self.center
        )

    def __repr__(self):
        return str(self)



# Copyright (C) 2021 torque <torque@users.noreply.github.com>

# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.

# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import annotations
from typing import Optional

cimport cython

import enum

from . cimport binarize as bnz, qrdecdecl
from . import datatypes, image

from libc.string cimport memcpy
from libc.stdio cimport printf, fopen, fwrite, fclose, FILE


QREccLevelMap = {
    0: datatypes.QREccLevel.L,
    1: datatypes.QREccLevel.M,
    2: datatypes.QREccLevel.Q,
    3: datatypes.QREccLevel.H,
}

cdef class QRdecDecoder:
    """
    QR code decoder using the QRdec backend.

    .. note::

        This decoder requires that input images have inverted reflectance in
        order to be able to decode them. The binarization defaults
        on :meth:`decode` handle performing this inversion, but if either
        binarization or the inversion is disabled, images will not be decoded
        properly without preprocessing.

    :raises MemoryError: if the QR code reader context allocation fails.
    """

    cdef qrdecdecl.qr_reader *_chndl

    def __cinit__(self):
        self._chndl = qrdecdecl.qr_reader_alloc()
        if self._chndl is NULL:
            raise MemoryError

    def __dealloc__(self):
        if self._chndl is not NULL:
            qrdecdecl.qr_reader_free(self._chndl)

    def decode(
        self,
        image_data,
        binarize: bool = True,
        binarize_invert: bool = True,
        convert_data: bool = True,
        byte_charset: Optional[str] = "utf-8"
    ) -> list[datatypes.QRCode]:
        """
        Decode all detectable QR codes in an image.

        .. warning::

            Binarization is done in-place on the image data buffer and
            changes the data. While this operation should be idempotent, it
            mutates the input data, which is a side effect.

            Additionally, inverting during binarization is also done in-place on
            the image data buffer and is *not* idempotent, so if image data is
            reused, it will be re-inverted and not decode correctly the second
            time.

        :param image_data:
            A python object containing the pixel data of the image to search for
            QR codes. This can be any format supported
            by :class:`image.ImageLoader`.

        :param binarize:
            If ``True``, binarize the input image (i.e. convert all
            pixels to either fully black or fully white). The decoder is
            unlikely to work properly on images that have not been binarized.

        :param binarize_invert:
            If ``True``, binarization inverts the reflectance (i.e dark pixels
            become white and light pixels become black).

        :param convert_data:
            If ``True``, all data entry payloads are  decoded into
            "native" python types:

            - :py:obj:`datatypes.QRDataType.NUMERIC` => :py:class:`int`,
            - :py:obj:`datatypes.QRDataType.ALPHANUMERIC` =>
              :py:class:`str` as ``ascii``
            - :py:obj:`datatypes.QRDataType.KANJI` =>
              :py:class:`str` as ``shift-jis``
            - :py:obj:`datatypes.QRDataType.BYTE` =>
              :py:class:`str` using the ``byte_charset`` parameter

            If ``False``, no conversion will be done and all data entry data
            will be returned as :py:class:`bytes`.

        :param byte_charset:
            The charset to use for converting all decoded data entries of type
            :py:obj:`datatypes.QRDataType.BYTE` found in the image. If
            ``None``, then data entries of type
            :py:obj:`datatypes.QRDataType.BYTE` will not be converted,
            even if ``convert_data`` is ``True``.

        :return: A list of objects containing information from the decoded QR
            codes.

        :raises TypeError: if ``image_data`` is of an unsupported type.
        :raises ValueError: if ``image_data`` is malformed somehow.
        """
        cdef qrdecdecl.qr_code_data_list qrlist

        if not isinstance(image_data, image.ImageLoader):
            image_data = image.ImageLoader(image_data)

        qrdecdecl.qr_code_data_list_init(&qrlist)

        cdef qrdecdecl.uint8[::1] imagebytes = image_data.data
        cdef int idx = 0
        cdef qrdecdecl.qr_code_data *code
        decoded: list[datatypes.QRCode] = []

        if binarize:
            bnz.binarize(
                image_data.data,
                image_data.width,
                image_data.height,
                binarize_invert,
            )

        if convert_data:
            if byte_charset is not None:
                byte_converter = lambda d: d.decode(byte_charset)
            else:
                byte_converter = lambda d: d

            converters = {
                datatypes.QRDataType.NUMERIC: int,
                datatypes.QRDataType.ALPHANUMERIC: lambda d: d.decode("ascii"),
                datatypes.QRDataType.KANJI: lambda d: d.decode("shift-jis"),
                datatypes.QRDataType.BYTE: byte_converter,
            }

        try:
            for idx in range(
                qrdecdecl.qr_reader_locate(
                    self._chndl,
                    &qrlist,
                    &imagebytes[0],
                    image_data.width,
                    image_data.height,
                )
            ):
                code = qrlist.qrdata + idx
                corners = tuple(
                    (code.bbox[idx][0], code.bbox[idx][1])
                    for idx in (0, 1, 3, 2)
                )

                data_entries = tuple(
                    datatypes.QRCodeData(
                        entry.mode,
                        entry.payload.data.buf[:entry.payload.data.len],
                    )
                    for entry in code.entries[:code.nentries]
                    if qrdecdecl.QR_MODE_HAS_DATA(entry.mode)
                )

                if convert_data:
                    for entry in data_entries:
                        entry.data = converters[entry.type](entry.data)

                decoded.append(
                    datatypes.QRCode(
                        version=code.version,
                        ecc_level=QREccLevelMap[code.ecc_level],
                        mask=code.mask,
                        data_entries=data_entries,
                        corners=corners,
                        center=(code.center[0], code.center[1])
                    )
                )
        finally:
            qrdecdecl.qr_code_data_list_clear(&qrlist)

        return decoded
